/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "quakedef.h"

int gl_max_texture_size = 0;

cvar_t		gl_nobind = {"gl_nobind", "0"};

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;

int			translate_texture;
int			char_texture;

typedef struct
{
	int		texnum;
	float	sl, tl, sh, th;
} glpic_t;

byte		conback_buffer[sizeof (qpic_t) + sizeof (glpic_t)];
qpic_t		*conback = (qpic_t *) &conback_buffer;

int		gl_lightmap_format = 4;
int		gl_solid_format = 3;
int		gl_alpha_format = 4;

int		gl_filter_min = GL_LINEAR_MIPMAP_LINEAR;
int		gl_filter_max = GL_LINEAR;


int		texels;

gltexture_t	gltextures[MAX_GLTEXTURES];
int			numgltextures;


/*
=============================================================================

scrap allocation

Allocate all the little status bar obejcts into a single texture
to crutch up stupid hardware / drivers

=============================================================================
*/

#define	MAX_SCRAPS		2
#define	BLOCK_WIDTH		256
#define	BLOCK_HEIGHT	256

int			scrap_allocated[MAX_SCRAPS][BLOCK_WIDTH];
byte		scrap_texels[MAX_SCRAPS][BLOCK_WIDTH*BLOCK_HEIGHT * 4];
qboolean	scrap_dirty;
int			scrap_texnum;

// returns a texture number and the position inside it
int Scrap_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		bestx;
	int		texnum;

	for (texnum = 0; texnum < MAX_SCRAPS; texnum++)
	{
		best = BLOCK_HEIGHT;

		for (i = 0; i < BLOCK_WIDTH - w; i++)
		{
			best2 = 0;

			for (j = 0; j<w; j++)
			{
				if (scrap_allocated[texnum][i + j] >= best)
					break;
				if (scrap_allocated[texnum][i + j] > best2)
					best2 = scrap_allocated[texnum][i + j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT)
			continue;

		for (i = 0; i < w; i++)
			scrap_allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("Scrap_AllocBlock: full");
	return 0;
}

int	scrap_uploads;

void Scrap_Upload (void)
{
	int		texnum;

	scrap_uploads++;

	for (texnum = 0; texnum < MAX_SCRAPS; texnum++)
	{
		GL_BindTexture (0, scrap_texnum + texnum);
		GL_Upload8 (scrap_texels[texnum], d_8to24table_rgba, BLOCK_WIDTH, BLOCK_HEIGHT, TEX_ALPHA);
	}
	scrap_dirty = false;
}

//=============================================================================
/* Support Routines */

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

int		pic_texels;
int		pic_count;

qpic_t *Draw_PicFromWad (char *name)
{
	qpic_t	*p;
	glpic_t	*gl;

	p = W_GetLumpName (name);
	gl = (glpic_t *) p->data;

	// load little ones into the scrap
	if (p->width < 64 && p->height < 64)
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

		texnum = Scrap_AllocBlock (p->width, p->height, &x, &y);
		scrap_dirty = true;
		k = 0;
		for (i = 0; i < p->height; i++)
			for (j = 0; j < p->width; j++, k++)
				scrap_texels[texnum][(y + i)*BLOCK_WIDTH + x + j] = p->data[k];
		texnum += scrap_texnum;
		gl->texnum = texnum;
		gl->sl = (x + 0.01) / (float) BLOCK_WIDTH;
		gl->sh = (x + p->width - 0.01) / (float) BLOCK_WIDTH;
		gl->tl = (y + 0.01) / (float) BLOCK_WIDTH;
		gl->th = (y + p->height - 0.01) / (float) BLOCK_WIDTH;

		pic_count++;
		pic_texels += p->width*p->height;
	}
	else
	{
		gl->texnum = GL_LoadPicTexture (p);
		gl->sl = 0;
		gl->sh = 1;
		gl->tl = 0;
		gl->th = 1;
	}
	return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
	glpic_t		*gl;

	for (pic = menu_cachepics, i = 0; i < menu_numcachepics; pic++, i++)
		if (!strcmp (path, pic->name))
			return &pic->pic;

	if (menu_numcachepics == MAX_CACHED_PICS)
		Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");
	menu_numcachepics++;
	strcpy (pic->name, path);

	// load the pic from disk
	dat = (qpic_t *) COM_LoadTempFile (path);
	if (!dat)
		Sys_Error ("Draw_CachePic: failed to load %s", path);
	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
		memcpy (menuplyr_pixels, dat->data, dat->width*dat->height);

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

	gl = (glpic_t *) pic->pic.data;
	gl->texnum = GL_LoadPicTexture (dat);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	return &pic->pic;
}


void Draw_CharToConback (int num, byte *dest)
{
	int		row, col;
	byte	*source;
	int		drawline;
	int		x;

	row = num >> 4;
	col = num & 15;
	source = draw_chars + (row << 10) + (col << 3);

	drawline = 8;

	while (drawline--)
	{
		for (x = 0; x < 8; x++)
			if (source[x] != 255)
				dest[x] = 0x60 + source[x];
		source += 128;
		dest += 320;
	}

}

typedef struct
{
	char *name;
	int	minimize, maximize;
} glmode_t;

glmode_t modes[] = {
	{"GL_NEAREST", GL_NEAREST, GL_NEAREST},
	{"GL_LINEAR", GL_LINEAR, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_NEAREST", GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_NEAREST", GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_LINEAR", GL_NEAREST_MIPMAP_LINEAR, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_LINEAR", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR}
};

/*
===============
Draw_TextureMode_f
===============
*/
void Draw_TextureMode_f (void)
{
	int		i;
	gltexture_t	*glt;

	if (Cmd_Argc () == 1)
	{
		for (i = 0; i < 6; i++)
			if (gl_filter_min == modes[i].minimize)
			{
				Con_Printf ("%s\n", modes[i].name);
				return;
			}
		Con_Printf ("current filter is unknown???\n");
		return;
	}

	for (i = 0; i < 6; i++)
	{
		if (!Q_strcasecmp (modes[i].name, Cmd_Argv (1)))
			break;
	}
	if (i == 6)
	{
		Con_Printf ("bad filter name\n");
		return;
	}

	gl_filter_min = modes[i].minimize;
	gl_filter_max = modes[i].maximize;

	// change all the existing mipmap texture objects
	for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
	{
		if (glt->flags & TEX_MIPMAP)
		{
			GL_BindTexture (0, glt->texnum);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
		}
	}
}

/*
===============
Draw_Init
===============
*/
void Draw_Init (void)
{
	int		i;
	qpic_t	*cb;
	byte	*dest, *src;
	int		x, y;
	char	ver[40];
	glpic_t	*gl;
	int		start;
	byte	*ncdata;
	int		f, fstep;

	Cvar_RegisterVariable (&gl_nobind);
	Cmd_AddCommand ("gl_texturemode", &Draw_TextureMode_f);

	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_max_texture_size);

	// load the console background and the charset
	// by hand, because we need to write the version
	// string into the background before turning
	// it into a texture
	draw_chars = W_GetLumpName ("conchars");
	for (i = 0; i < 256 * 64; i++)
		if (draw_chars[i] == 0)
			draw_chars[i] = 255;	// proper transparent color

	// now turn them into textures
	char_texture = GL_LoadTexture ("charset", 128, 128, draw_chars, d_8to24table_rgba, TEX_ALPHA);

	start = Hunk_LowMark ();

	cb = (qpic_t *) COM_LoadTempFile ("gfx/conback.lmp");
	if (!cb)
		Sys_Error ("Couldn't load gfx/conback.lmp");
	SwapPic (cb);

	// hack the version number directly into the pic
#if defined(__linux__)
	sprintf (ver, "(Linux %2.2f, gl %4.2f) %4.2f", (float)LINUX_VERSION, (float)GLQUAKE_VERSION, (float)VERSION);
#else
	sprintf (ver, "(gl %4.2f) %4.2f", (float) GLQUAKE_VERSION, (float) VERSION);
#endif
	dest = cb->data + 320 * 186 + 320 - 11 - 8 * strlen (ver);
	y = strlen (ver);
	for (x = 0; x < y; x++)
		Draw_CharToConback (ver[x], dest + (x << 3));

#if 0
	conback->width = vid.conwidth;
	conback->height = vid.conheight;

	// scale console to vid size
	dest = ncdata = Hunk_AllocName(vid.conwidth * vid.conheight, "conback");

	for (y=0 ; y<vid.conheight ; y++, dest += vid.conwidth)
	{
		src = cb->data + cb->width * (y*cb->height/vid.conheight);
		if (vid.conwidth == cb->width)
			memcpy (dest, src, vid.conwidth);
		else
		{
			f = 0;
			fstep = cb->width*0x10000/vid.conwidth;
			for (x=0 ; x<vid.conwidth ; x+=4)
			{
				dest[x] = src[f>>16];
				f += fstep;
				dest[x+1] = src[f>>16];
				f += fstep;
				dest[x+2] = src[f>>16];
				f += fstep;
				dest[x+3] = src[f>>16];
				f += fstep;
			}
		}
	}
#else
	conback->width = cb->width;
	conback->height = cb->height;
	ncdata = cb->data;
#endif

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	gl = (glpic_t *) conback->data;
	gl->texnum = GL_LoadTexture ("conback", conback->width, conback->height, ncdata, d_8to24table_rgba, 0);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;
	conback->width = vid.width;
	conback->height = vid.height;

	// free loaded console
	Hunk_FreeToLowMark (start);

	// save a texture slot for translated picture
	translate_texture = texture_extension_number++;

	// save slots for scraps
	scrap_texnum = texture_extension_number;
	texture_extension_number += MAX_SCRAPS;

	// get the other pics we need
	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");
}



/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
	byte			*dest;
	byte			*source;
	unsigned short	*pusdest;
	int				drawline;
	int				row, col;
	float			frow, fcol, size;

	if (num == 32)
		return;		// space

	num &= 255;

	if (y <= -8)
		return;			// totally off screen

	row = num >> 4;
	col = num & 15;

	frow = row*0.0625;
	fcol = col*0.0625;
	size = 0.0625;

	GL_BindTexture (0, char_texture);
	GL_BindPrograms (r_drawtexturedvp, r_drawtexturedfp);

	glBegin (GL_QUADS);
	glTexCoord2f (fcol, frow);
	glVertex2f (x, y);
	glTexCoord2f (fcol + size, frow);
	glVertex2f (x + 8, y);
	glTexCoord2f (fcol + size, frow + size);
	glVertex2f (x + 8, y + 8);
	glTexCoord2f (fcol, frow + size);
	glVertex2f (x, y + 8);
	glEnd ();
}

/*
================
Draw_String
================
*/
void Draw_String (int x, int y, char *str)
{
	while (*str)
	{
		Draw_Character (x, y, *str);
		str++;
		x += 8;
	}
}

/*
================
Draw_DebugChar

Draws a single character directly to the upper right corner of the screen.
This is for debugging lockups by drawing different chars in different parts
of the code.
================
*/
void Draw_DebugChar (char num)
{
}

/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
	byte			*dest, *source;
	unsigned short	*pusdest;
	int				v, u;
	glpic_t			*gl;

	if (scrap_dirty)
		Scrap_Upload ();
	gl = (glpic_t *) pic->data;

	glColor4f (1, 1, 1, alpha);

	GL_BindPrograms (r_drawtexturedvp, r_drawtexturedfp);
	GL_BindTexture (0, gl->texnum);

	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x + pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x + pic->width, y + pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y + pic->height);
	glEnd ();

	glColor4f (1, 1, 1, 1);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	byte			*dest, *source;
	unsigned short	*pusdest;
	int				v, u;
	glpic_t			*gl;

	if (scrap_dirty)
		Scrap_Upload ();
	gl = (glpic_t *) pic->data;
	glColor4f (1, 1, 1, 1);

	GL_BindPrograms (r_drawtexturedvp, r_drawtexturedfp);

	GL_BindTexture (0, gl->texnum);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x + pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x + pic->width, y + pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y + pic->height);
	glEnd ();
}


/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
	byte	*dest, *source, tbyte;
	unsigned short	*pusdest;
	int				v, u;

	if (x < 0 || (unsigned) (x + pic->width) > vid.width || y < 0 ||
		(unsigned) (y + pic->height) > vid.height)
	{
		Sys_Error ("Draw_TransPic: bad coordinates");
	}

	Draw_Pic (x, y, pic);
}


/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	int				v, u, c;
	unsigned		trans[64 * 64], *dest;
	byte			*src;
	int				p;

	GL_BindTexture (0, translate_texture);

	c = pic->width * pic->height;

	dest = trans;
	for (v = 0; v < 64; v++, dest += 64)
	{
		src = &menuplyr_pixels[((v*pic->height) >> 6) *pic->width];
		for (u = 0; u < 64; u++)
		{
			p = src[(u*pic->width) >> 6];
			if (p == 255)
				dest[u] = p;
			else
				dest[u] = d_8to24table_rgba[translation[p]];
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	GL_BindPrograms (r_drawtexturedvp, r_drawtexturedfp);

	glColor3f (1, 1, 1);
	glBegin (GL_QUADS);
	glTexCoord2f (0, 0);
	glVertex2f (x, y);
	glTexCoord2f (1, 0);
	glVertex2f (x + pic->width, y);
	glTexCoord2f (1, 1);
	glVertex2f (x + pic->width, y + pic->height);
	glTexCoord2f (0, 1);
	glVertex2f (x, y + pic->height);
	glEnd ();
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	int y = (vid.height * 3) >> 2;

	if (lines > y)
		Draw_Pic (0, lines - vid.height, conback);
	else
		Draw_AlphaPic (0, lines - vid.height, conback, (float) (1.2 * lines) / y);
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
	GL_BindPrograms (r_drawtexturedvp, r_drawtexturedfp);

	glColor3f (1, 1, 1);
	GL_BindTexture (0, *(int *) draw_backtile->data);
	glBegin (GL_QUADS);
	glTexCoord2f (x / 64.0, y / 64.0);
	glVertex2f (x, y);
	glTexCoord2f ((x + w) / 64.0, y / 64.0);
	glVertex2f (x + w, y);
	glTexCoord2f ((x + w) / 64.0, (y + h) / 64.0);
	glVertex2f (x + w, y + h);
	glTexCoord2f (x / 64.0, (y + h) / 64.0);
	glVertex2f (x, y + h);
	glEnd ();
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	GL_BindPrograms (r_drawcolouredvp, r_drawcolouredfp);

	glDisable (GL_TEXTURE_2D);
	glColor3f (host_basepal[c * 3] / 255.0,
		host_basepal[c * 3 + 1] / 255.0,
		host_basepal[c * 3 + 2] / 255.0);

	glBegin (GL_QUADS);

	glVertex2f (x, y);
	glVertex2f (x + w, y);
	glVertex2f (x + w, y + h);
	glVertex2f (x, y + h);

	glEnd ();
	glColor3f (1, 1, 1);
	glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	GL_BindPrograms (r_drawcolouredvp, r_drawcolouredfp);

	glColor4f (0, 0, 0, 0.5);
	glBegin (GL_QUADS);

	glVertex2f (0, 0);
	glVertex2f (vid.width, 0);
	glVertex2f (vid.width, vid.height);
	glVertex2f (0, vid.height);

	glEnd ();
	glColor4f (1, 1, 1, 1);

	Sbar_Changed ();
}

//=============================================================================

/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
}


/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{
}

/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void GL_Set2D (void)
{
	glViewport (0, 0, vid.width, vid.height);

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glOrtho (0, vid.width, vid.height, 0, -99999, 99999);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glDisable (GL_CULL_FACE);

	GL_BlendState (BS_ALPHABLEND);
	GL_DepthState (DS_NONE);

	glColor4f (1, 1, 1, 1);
}

//====================================================================

/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	int		i;
	gltexture_t	*glt;

	for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
	{
		if (!strcmp (identifier, glt->identifier))
			return gltextures[i].texnum;
	}

	return -1;
}


static const float miptexel[256] = {
	0.000000f, 0.000015f, 0.000062f, 0.000138f, 0.000246f, 0.000384f, 0.000554f, 0.000754f, 0.000984f, 0.001246f, 0.001538f, 0.001861f, 0.002215f, 0.002599f, 0.003014f, 0.003460f,
	0.003937f, 0.004444f, 0.004983f, 0.005552f, 0.006151f, 0.006782f, 0.007443f, 0.008135f, 0.008858f, 0.009612f, 0.010396f, 0.011211f, 0.012057f, 0.012933f, 0.013841f, 0.014779f,
	0.015748f, 0.016747f, 0.017778f, 0.018839f, 0.019931f, 0.021053f, 0.022207f, 0.023391f, 0.024606f, 0.025852f, 0.027128f, 0.028435f, 0.029773f, 0.031142f, 0.032541f, 0.033972f,
	0.035433f, 0.036924f, 0.038447f, 0.040000f, 0.041584f, 0.043199f, 0.044844f, 0.046521f, 0.048228f, 0.049965f, 0.051734f, 0.053533f, 0.055363f, 0.057224f, 0.059116f, 0.061038f,
	0.062991f, 0.064975f, 0.066990f, 0.069035f, 0.071111f, 0.073218f, 0.075356f, 0.077524f, 0.079723f, 0.081953f, 0.084214f, 0.086505f, 0.088827f, 0.091180f, 0.093564f, 0.095978f,
	0.098424f, 0.100900f, 0.103406f, 0.105944f, 0.108512f, 0.111111f, 0.113741f, 0.116401f, 0.119093f, 0.121815f, 0.124567f, 0.127351f, 0.130165f, 0.133010f, 0.135886f, 0.138793f,
	0.141730f, 0.144698f, 0.147697f, 0.150727f, 0.153787f, 0.156878f, 0.160000f, 0.163153f, 0.166336f, 0.169550f, 0.172795f, 0.176071f, 0.179377f, 0.182714f, 0.186082f, 0.189481f,
	0.192910f, 0.196371f, 0.199862f, 0.203383f, 0.206936f, 0.210519f, 0.214133f, 0.217778f, 0.221453f, 0.225160f, 0.228897f, 0.232664f, 0.236463f, 0.240292f, 0.244152f, 0.248043f,
	0.251965f, 0.255917f, 0.259900f, 0.263914f, 0.267959f, 0.272034f, 0.276140f, 0.280277f, 0.284444f, 0.288643f, 0.292872f, 0.297132f, 0.301423f, 0.305744f, 0.310096f, 0.314479f,
	0.318893f, 0.323337f, 0.327812f, 0.332318f, 0.336855f, 0.341423f, 0.346021f, 0.350650f, 0.355310f, 0.360000f, 0.364721f, 0.369473f, 0.374256f, 0.379070f, 0.383914f, 0.388789f,
	0.393695f, 0.398631f, 0.403599f, 0.408597f, 0.413626f, 0.418685f, 0.423775f, 0.428897f, 0.434048f, 0.439231f, 0.444444f, 0.449689f, 0.454964f, 0.460269f, 0.465606f, 0.470973f,
	0.476371f, 0.481799f, 0.487259f, 0.492749f, 0.498270f, 0.503822f, 0.509404f, 0.515017f, 0.520661f, 0.526336f, 0.532042f, 0.537778f, 0.543545f, 0.549343f, 0.555171f, 0.561030f,
	0.566920f, 0.572841f, 0.578793f, 0.584775f, 0.590788f, 0.596832f, 0.602907f, 0.609012f, 0.615148f, 0.621315f, 0.627513f, 0.633741f, 0.640000f, 0.646290f, 0.652611f, 0.658962f,
	0.665344f, 0.671757f, 0.678201f, 0.684675f, 0.691180f, 0.697716f, 0.704283f, 0.710880f, 0.717509f, 0.724168f, 0.730857f, 0.737578f, 0.744329f, 0.751111f, 0.757924f, 0.764767f,
	0.771642f, 0.778547f, 0.785483f, 0.792449f, 0.799446f, 0.806474f, 0.813533f, 0.820623f, 0.827743f, 0.834894f, 0.842076f, 0.849289f, 0.856532f, 0.863806f, 0.871111f, 0.878447f,
	0.885813f, 0.893210f, 0.900638f, 0.908097f, 0.915586f, 0.923106f, 0.930657f, 0.938239f, 0.945852f, 0.953495f, 0.961169f, 0.968874f, 0.976609f, 0.984375f, 0.992172f, 1.000000f
};


int AverageMip (int _1, int _2, int _3, int _4)
{
	return (_1 + _2 + _3 + _4) >> 2;
}


int AverageMipGC (int _1, int _2, int _3, int _4)
{
	return (int) (255.0f * sqrt ((miptexel[_1] + miptexel[_2] + miptexel[_3] + miptexel[_4]) / 4.0f));
}


unsigned *R_ResampleTexture (unsigned *in, int inwidth, int inheight, int outwidth, int outheight)
{
	// allocating the out buffer before the pool mark so that we can return it
	unsigned *out = (unsigned *) Hunk_Alloc (outwidth * outheight * 4);

	// and now get the mark because all allocations after this will be released
	int mark = Hunk_LowMark ();

	unsigned *p1 = (unsigned *) Hunk_Alloc (outwidth * 4);
	unsigned *p2 = (unsigned *) Hunk_Alloc (outwidth * 4);

	unsigned fracstep = inwidth * 0x10000 / outwidth;
	unsigned frac = fracstep >> 2;

	int i, j;

	for (i = 0; i < outwidth; i++)
	{
		p1[i] = 4 * (frac >> 16);
		frac += fracstep;
	}

	frac = 3 * (fracstep >> 2);

	for (i = 0; i < outwidth; i++)
	{
		p2[i] = 4 * (frac >> 16);
		frac += fracstep;
	}

	for (i = 0; i < outheight; i++)
	{
		unsigned *outrow = out + (i * outwidth);
		unsigned *inrow0 = in + inwidth * (int) (((i + 0.25f) * inheight) / outheight);
		unsigned *inrow1 = in + inwidth * (int) (((i + 0.75f) * inheight) / outheight);

		for (j = 0; j < outwidth; j++)
		{
			byte *pix1 = (byte *) inrow0 + p1[j];
			byte *pix2 = (byte *) inrow0 + p2[j];
			byte *pix3 = (byte *) inrow1 + p1[j];
			byte *pix4 = (byte *) inrow1 + p2[j];

			// don't gamma correct the alpha channel
			((byte *) &outrow[j])[0] = AverageMipGC (pix1[0], pix2[0], pix3[0], pix4[0]);
			((byte *) &outrow[j])[1] = AverageMipGC (pix1[1], pix2[1], pix3[1], pix4[1]);
			((byte *) &outrow[j])[2] = AverageMipGC (pix1[2], pix2[2], pix3[2], pix4[2]);
			((byte *) &outrow[j])[3] = AverageMip (pix1[3], pix2[3], pix3[3], pix4[3]);
		}
	}

	Hunk_FreeToLowMark (mark);

	return out;
}


unsigned *R_MipMap (unsigned *data, int width, int height)
{
	// because each SRD must have it's own data we can't mipmap in-place otherwise we'll corrupt the previous miplevel
	unsigned *trans = (unsigned *) Hunk_Alloc ((width >> 1) * (height >> 1) * 4);
	byte *in = (byte *) data;
	byte *out = (byte *) trans;

	int i, j;

	// do this after otherwise it will interfere with the allocation size above
	width <<= 2;
	height >>= 1;

	for (i = 0; i < height; i++, in += width)
	{
		for (j = 0; j < width; j += 8, out += 4, in += 8)
		{
			// don't gamma correct the alpha channel
			out[0] = AverageMipGC (in[0], in[4], in[width + 0], in[width + 4]);
			out[1] = AverageMipGC (in[1], in[5], in[width + 1], in[width + 5]);
			out[2] = AverageMipGC (in[2], in[6], in[width + 2], in[width + 6]);
			out[3] = AverageMip (in[3], in[7], in[width + 3], in[width + 7]);
		}
	}

	return trans;
}


void R_AlphaEdgeFix (byte *data, int width, int height)
{
	byte *dest = data;

	int i, j;

	for (i = 0; i < height; i++)
	{
		int lastrow = width * 4 * ((i == 0) ? height - 1 : i - 1);
		int thisrow = width * 4 * i;
		int nextrow = width * 4 * ((i == height - 1) ? 0 : i + 1);

		for (j = 0; j < width; j++, dest += 4)
		{
			// not transparent
			if (!dest[3])
			{
				int lastpix = 4 * ((j == 0) ? width - 1 : j - 1);
				int thispix = 4 * j;
				int nextpix = 4 * ((j == width - 1) ? 0 : j + 1);
				int n = 0, b, c[3] = {0, 0, 0};

				b = lastrow + lastpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = thisrow + lastpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = nextrow + lastpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = lastrow + thispix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = nextrow + thispix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = lastrow + nextpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = thisrow + nextpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }
				b = nextrow + nextpix; if (data[b + 3]) { c[0] += data[b]; c[1] += data[b + 1]; c[2] += data[b + 2]; n++; }

				// average all non-transparent neighbors
				if (n)
				{
					dest[0] = (byte) (c[0] / n);
					dest[1] = (byte) (c[1] / n);
					dest[2] = (byte) (c[2] / n);
				}
			}
		}
	}
}


void GL_GenerateTextureMipmaps (unsigned *data, int width, int height, int miplevel, qboolean mipmap)
{
	// lower-bound
	if (width < 1) width = 1;
	if (height < 1) height = 1;

	// upload this level
	glTexImage2D (GL_TEXTURE_2D, miplevel, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

	// upload the next level (if mipmapped)
	if ((width > 1 || height > 1) && mipmap)
	{
		// choose the appropriate filter
		if ((width & 1) || (height & 1))
			data = R_ResampleTexture (data, width, height, width > 1 ? width >> 1 : 1, height > 1 ? height >> 1 : 1);
		else data = R_MipMap (data, width, height);

		GL_GenerateTextureMipmaps (data, width >> 1, height >> 1, miplevel + 1, mipmap);
	}
}


void GL_Upload8 (byte *data, unsigned *palette, int width, int height, int flags)
{
	int			mark = Hunk_LowMark ();
	unsigned	*trans = (unsigned *) Hunk_Alloc (width * height * 4);
	int			scaled_width, scaled_height;
	int			s = width * height;
	int			i;

	// expand to 32-bit; s & 3 happens on marcher so just do it un-batched
	for (i = 0; i < s; i++)
		trans[i] = palette[data[i]];

	// run the alpha fixup before anything else
	if (flags & TEX_ALPHA) R_AlphaEdgeFix ((byte *) trans, width, height);

	if (GLEW_ARB_texture_non_power_of_two)
	{
		scaled_width = width;
		scaled_height = height;
	}
	else
	{
		for (scaled_width = 1; scaled_width < width; scaled_width <<= 1);
		for (scaled_height = 1; scaled_height < height; scaled_height <<= 1);
	}

	if (scaled_width > gl_max_texture_size) scaled_width = gl_max_texture_size;
	if (scaled_height > gl_max_texture_size) scaled_height = gl_max_texture_size;

	if (scaled_width != width || scaled_height != height)
		trans = R_ResampleTexture (trans, width, height, scaled_width, scaled_height);

	texels += scaled_width * scaled_height;

	// generate mipmaps beginning at level 0
	GL_GenerateTextureMipmaps (trans, scaled_width, scaled_height, 0, (flags & TEX_MIPMAP));

	if (flags & TEX_MIPMAP)
	{
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}
	else
	{
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_max);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}

	Hunk_FreeToLowMark (mark);
}


/*
================
GL_LoadTexture
================
*/
int GL_LoadTexture (char *identifier, int width, int height, byte *data, unsigned *palette, int flags)
{
	int	i;
	gltexture_t	*glt;
	unsigned short crc = CRC_Block (data, width * height);

	// see if the texture is allready present
	if (identifier[0])
	{
		// HACK HACK HACK -- taken from tomazquake
		if (strstr (identifier, "shot1sid") && width == 32 && height == 32 && crc == 65393)
		{
			// This texture in b_shell1.bsp has some of the first 32 pixels painted white.
			// They are invisible in software, but look really ugly in GL. So we just copy
			// 32 pixels from the bottom to make it look nice.
			memcpy (data, data + 32 * 31, 32);
		}

		for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		{
			if (width != glt->width) continue;
			if (height != glt->height) continue;
			if (flags != glt->flags) continue;
			if (crc != glt->crc) continue;
			if (palette != glt->palette) continue;
			if (strcmp (identifier, glt->identifier)) continue;

			// mark as used
			glt->registration_sequence = r_registration_sequence;

			return glt->texnum;
		}
	}

	glt = &gltextures[numgltextures];
	numgltextures++;

	glt->texnum = texture_extension_number;
	texture_extension_number++;

	strcpy (glt->identifier, identifier);
	glt->width = width;
	glt->height = height;
	glt->flags = flags;
	glt->crc = crc;
	glt->palette = palette;
	glt->registration_sequence = r_registration_sequence;

	GL_BindTexture (0, glt->texnum);
	GL_Upload8 (data, palette, width, height, flags);

	return glt->texnum;
}

/*
================
GL_LoadPicTexture
================
*/
int GL_LoadPicTexture (qpic_t *pic)
{
	return GL_LoadTexture ("", pic->width, pic->height, pic->data, d_8to24table_rgba, TEX_ALPHA);
}


void GL_FreeUnusedTextures (void)
{
}


