/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// disable data conversion warnings

#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA
  
#ifdef _WIN32
#include <windows.h>
#endif

#include "glew.h"
#include <GL/gl.h>
#include <GL/glu.h>

qboolean GL_BeginRendering (void);
void GL_EndRendering (void);

extern int r_registration_sequence;

extern	int texture_extension_number;
extern	float	gldepthmin, gldepthmax;

// textures
typedef struct gltexture_s
{
	int		texnum;
	char	identifier[64];
	int		width, height;
	int		flags;
	unsigned short crc;
	unsigned *palette;
	int registration_sequence;
} gltexture_t;

#define	MAX_GLTEXTURES	1024

#define TEX_MIPMAP		(1 << 0)
#define TEX_ALPHA		(1 << 1)
#define TEX_LUMA		(1 << 2)

void GL_Upload8 (byte *data, unsigned *palette, int width, int height, int flags);
int GL_LoadTexture (char *identifier, int width, int height, byte *data, unsigned *palette,  int flags);
int GL_FindTexture (char *identifier);
void GL_FreeUnusedTextures (void);

// r_local.h -- private refresh defs
#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
					// normalizing factor so player model works out to about
					//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		480

#define TILE_SIZE		128		// size of textures generated by R_GenTiledSurf

#define SKYSHIFT		7
#define	SKYSIZE			(1 << SKYSHIFT)
#define SKYMASK			(SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01


void R_ReadPointFile_f (void);
texture_t *R_TextureAnimation (texture_t *base);

// state
// blend states are exclusive of each other
#define BS_NONE			0
#define BS_ALPHABLEND	1
#define BS_ADDITIVE		2
#define BS_MAXBLEND		3

// depth states are combined with each other
#define DS_NONE				(0)
#define DS_WRITE			(1 << 0)
#define DS_TESTLESSEQUAL	(1 << 1)
#define DS_TESTEQUAL		(1 << 2)

void GL_BlendState (int bs);
void GL_DepthState (int ds);

// vertex attrib arrays are combined with each other
#define VAA0	(1 << 0)
#define VAA1	(1 << 1)
#define VAA2	(1 << 2)
#define VAA3	(1 << 3)
#define VAA4	(1 << 4)
#define VAA5	(1 << 5)
#define VAA6	(1 << 6)
#define VAA7	(1 << 7)
#define VAA8	(1 << 8)
#define VAA9	(1 << 9)
#define VAA10	(1 << 10)
#define VAA11	(1 << 11)
#define VAA12	(1 << 12)
#define VAA13	(1 << 13)
#define VAA14	(1 << 14)
#define VAA15	(1 << 15)

void GL_EnableVertexAttribArrays (int arrays);


// models
void R_DrawAliasModel (entity_t *e, GLuint vp, GLuint fp);
void R_DrawBrushModel (entity_t *e);
void R_DrawSpriteModel (entity_t *e);

//====================================================


extern	entity_t	r_worldentity;
extern	qboolean	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	entity_t	*currententity;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;
extern	int			r_dlightframecount;
extern	mplane_t	frustum[4];


// view origin (padded for shaders)
extern	float	vup[4];
extern	float	vpn[4];
extern	float	vright[4];
extern	float	r_origin[4];

// screen size info
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	texture_t	*r_notexture_mip;
extern	float		d_lightstylevalue[256];	// 8.8 fraction of base light value

extern	int r_spottexture;
extern	int	playertextures;

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_fullbright;
extern	cvar_t	r_lightmap;
extern	cvar_t	r_shadows;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;

extern	cvar_t	gl_cull;
extern	cvar_t	gl_poly;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_nocolors;

extern	int		gl_lightmap_format;
extern	int		gl_solid_format;
extern	int		gl_alpha_format;

extern	glmatrix	r_world_matrix;
extern	glmatrix	r_projection_matrix;
extern	glmatrix	r_mvp_matrix;

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;

qboolean R_CullSphere (float *center, float radius);
qboolean R_CullBox (float *emins, float *emaxs);

void R_TranslatePlayerSkin (int playernum);
void GL_BindTexture (int unit, int texnum);

// surfaces
typedef struct extrasurface_s
{
	// sky, water, fullbrights
	msurface_t *surf;
	entity_t *ent;
} extrasurface_t;

#define MAX_EXTRA_SURFACES 65536

extern extrasurface_t r_extrasurfaces[MAX_EXTRA_SURFACES];
extern int r_numextrasurfaces;

void R_DrawExtraSurfaces (void);
void R_DrawSkyChain (msurface_t *surf);
void R_DrawWaterChain (msurface_t *surf);
void R_DrawLumaChain (msurface_t *surf);

void R_SetupSkyShaderParams (void);
void R_SetupWarpShaderParams (void);

void GL_BuildSurfaces (void);
void R_RegisterSurface (msurface_t *surf);

void GL_BeginDrawingSurfaces (void);
void GL_DrawSurface (msurface_t *surf);
void GL_EndDrawingSurfaces (void);

void R_ClearTextureChains (model_t *mod);

// lights
typedef struct lfrag_s
{
	struct mleaf_s		*leaf;
	struct lfrag_s		*leafnext;
	struct staticlight_s *light;
	struct lfrag_s		*lightnext;
} lfrag_t;

typedef struct staticlight_s
{
	// matches entity_t struct in light.exe
	char	classname[64];
	vec3_t	origin;
	float	color[3];			// 0..1 range but may overbright
	float	angle;
	int		light;
	int		style;
	char	target[32];
	char	targetname[32];
	lfrag_t *lfrag;
	int		visframe;
	struct staticlight_s *targetent;
} staticlight_t;

typedef struct spotparms_s
{
	float spotvec[4]; // padded for shader
	float falloff;
	qboolean isSpotLight;
} spotparms_t;

#define DEFAULTLIGHTLEVEL	300

void R_LoadLights (void);
void R_RenderLightsAndShadows (void);
void R_DrawTextureChains_LightPass (model_t *mod, glmatrix *localmatrix);
void R_AddLfrags (staticlight_t *light);
void R_StoreLfrags (lfrag_t **pplfrag);


// shaders
extern GLuint r_depthpasssurfvp;
extern GLuint r_depthpassmeshvp;
extern GLuint r_depthpassfp;

extern GLuint r_drawtexturedvp;
extern GLuint r_drawtexturedfp;

extern GLuint r_drawcolouredvp;
extern GLuint r_drawcolouredfp;

extern GLuint r_fullbrightsurfvp;
extern GLuint r_fullbrightmeshvp;
extern GLuint r_fullbrightfp;

extern GLuint r_lightsurfvp;
extern GLuint r_lightmeshvp;

extern GLuint r_pointlightfp;
extern GLuint r_spotlightfp;

extern GLuint r_skysurfvp;
extern GLuint r_skysurffp;

extern GLuint r_warpsurfvp;
extern GLuint r_warpsurffp;

extern GLuint r_particlevp;
extern GLuint r_particlefp;

void GL_LoadShaders (void);

