/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// snd_mix.c -- portable code to mix sounds for snd_dma.c

#include "quakedef.h"

#ifdef _WIN32
#include "winquake.h"
#else
#define DWORD	unsigned long
#endif

#define	PAINTBUFFER_SIZE	512
portable_samplepair_t paintbuffer[PAINTBUFFER_SIZE];
int		snd_scaletable[32][256];
int 	*snd_p, snd_linear_count, snd_vol;
short	*snd_out;

__declspec (naked) void Snd_WriteLinearBlastStereo16 (void)
{
	// this is asm just for perf
	__asm
	{
		push esi
			push edi
			push ebx

			mov ecx, ds : dword ptr[snd_linear_count]
			mov ebx, ds : dword ptr[snd_p]
			mov esi, ds : dword ptr[snd_vol]
			mov edi, ds : dword ptr[snd_out]

LWLBLoopTop :
			mov eax, ds : dword ptr[-8 + ebx + ecx * 4]
			imul eax, esi
			sar eax, 8
			cmp eax, 07FFFh
			jg LClampHigh
			cmp eax, 0FFFF8000h
			jnl LClampDone
			mov eax, 0FFFF8000h
			jmp LClampDone

LClampHigh :
		mov eax, 07FFFh

LClampDone :
		   mov edx, ds : dword ptr[-4 + ebx + ecx * 4]
		   imul edx, esi
		   sar edx, 8
		   cmp edx, 07FFFh
		   jg LClampHigh2
		   cmp edx, 0FFFF8000h
		   jnl LClampDone2
		   mov edx, 0FFFF8000h
		   jmp LClampDone2

LClampHigh2 :
		mov edx, 07FFFh

LClampDone2 :
			shl edx, 16
			and eax, 0FFFFh
			or edx, eax
			mov ds : dword ptr[-4 + edi + ecx * 2], edx
			sub ecx, 2
			jnz LWLBLoopTop

			pop ebx
			pop edi
			pop esi
			ret
	}
}


void S_TransferStereo16 (int endtime)
{
	int		lpos;
	int		lpaintedtime;
	DWORD	*pbuf;
#ifdef _WIN32
	int		reps;
	DWORD	dwSize, dwSize2;
	DWORD	*pbuf2;
	HRESULT	hresult;
#endif

	snd_vol = volume.value * 256;

	snd_p = (int *) paintbuffer;
	lpaintedtime = paintedtime;

#ifdef _WIN32
	if (pDSBuf)
	{
		reps = 0;

		while ((hresult = pDSBuf->lpVtbl->Lock (pDSBuf, 0, gSndBufSize, &pbuf, &dwSize,
			&pbuf2, &dwSize2, 0)) != DS_OK)
		{
			if (hresult != DSERR_BUFFERLOST)
			{
				Con_Printf ("S_TransferStereo16: DS::Lock Sound Buffer Failed\n");
				S_Shutdown ();
				S_Startup ();
				return;
			}

			if (++reps > 10000)
			{
				Con_Printf ("S_TransferStereo16: DS: couldn't restore buffer\n");
				S_Shutdown ();
				S_Startup ();
				return;
			}
		}
	}
	else
#endif
	{
		pbuf = (DWORD *) shm->buffer;
	}

	while (lpaintedtime < endtime)
	{
		// handle recirculating buffer issues
		lpos = lpaintedtime & ((shm->samples >> 1) - 1);

		snd_out = (short *) pbuf + (lpos << 1);

		snd_linear_count = (shm->samples >> 1) - lpos;
		if (lpaintedtime + snd_linear_count > endtime)
			snd_linear_count = endtime - lpaintedtime;

		snd_linear_count <<= 1;

		// write a linear blast of samples
		Snd_WriteLinearBlastStereo16 ();

		snd_p += snd_linear_count;
		lpaintedtime += (snd_linear_count >> 1);
	}

#ifdef _WIN32
	if (pDSBuf)
		pDSBuf->lpVtbl->Unlock (pDSBuf, pbuf, dwSize, NULL, 0);
#endif
}

void S_TransferPaintBuffer (int endtime)
{
	int 	out_idx;
	int 	count;
	int 	out_mask;
	int 	*p;
	int 	step;
	int		val;
	int		snd_vol;
	DWORD	*pbuf;
#ifdef _WIN32
	int		reps;
	DWORD	dwSize, dwSize2;
	DWORD	*pbuf2;
	HRESULT	hresult;
#endif

	if (shm->samplebits == 16 && shm->channels == 2)
	{
		S_TransferStereo16 (endtime);
		return;
	}

	p = (int *) paintbuffer;
	count = (endtime - paintedtime) * shm->channels;
	out_mask = shm->samples - 1;
	out_idx = paintedtime * shm->channels & out_mask;
	step = 3 - shm->channels;
	snd_vol = volume.value * 256;

#ifdef _WIN32
	if (pDSBuf)
	{
		reps = 0;

		while ((hresult = pDSBuf->lpVtbl->Lock (pDSBuf, 0, gSndBufSize, &pbuf, &dwSize,
			&pbuf2, &dwSize2, 0)) != DS_OK)
		{
			if (hresult != DSERR_BUFFERLOST)
			{
				Con_Printf ("S_TransferPaintBuffer: DS::Lock Sound Buffer Failed\n");
				S_Shutdown ();
				S_Startup ();
				return;
			}

			if (++reps > 10000)
			{
				Con_Printf ("S_TransferPaintBuffer: DS: couldn't restore buffer\n");
				S_Shutdown ();
				S_Startup ();
				return;
			}
		}
	}
	else
#endif
	{
		pbuf = (DWORD *) shm->buffer;
	}

	if (shm->samplebits == 16)
	{
		short *out = (short *) pbuf;
		while (count--)
		{
			val = (*p * snd_vol) >> 8;
			p += step;
			if (val > 0x7fff)
				val = 0x7fff;
			else if (val < (short) 0x8000)
				val = (short) 0x8000;
			out[out_idx] = val;
			out_idx = (out_idx + 1) & out_mask;
		}
	}
	else if (shm->samplebits == 8)
	{
		unsigned char *out = (unsigned char *) pbuf;
		while (count--)
		{
			val = (*p * snd_vol) >> 8;
			p += step;
			if (val > 0x7fff)
				val = 0x7fff;
			else if (val < (short) 0x8000)
				val = (short) 0x8000;
			out[out_idx] = (val >> 8) + 128;
			out_idx = (out_idx + 1) & out_mask;
		}
	}

#ifdef _WIN32
	if (pDSBuf)
	{
		DWORD dwNewpos, dwWrite;
		int il = paintedtime;
		int ir = endtime - paintedtime;

		ir += il;

		pDSBuf->lpVtbl->Unlock (pDSBuf, pbuf, dwSize, NULL, 0);

		pDSBuf->lpVtbl->GetCurrentPosition (pDSBuf, &dwNewpos, &dwWrite);

		//		if ((dwNewpos >= il) && (dwNewpos <= ir))
		//			Con_Printf("%d-%d p %d c\n", il, ir, dwNewpos);
	}
#endif
}


/*
===============================================================================

CHANNEL MIXING

===============================================================================
*/

__declspec (naked) void SND_PaintChannelFrom8 (channel_t *ch, sfxcache_t *sc, int count)
{
	// this is done as asm so that it (1) avoids the 8-bit mixing bug in the C code, and (2) gets the perf back
	__asm
	{
		push esi
			push edi
			push ebx
			push ebp

			mov ebx, ds : dword ptr[4 + 16 + esp]
			mov esi, ds : dword ptr[8 + 16 + esp]
			mov eax, ds : dword ptr[4 + ebx]
			mov edx, ds : dword ptr[8 + ebx]
			cmp eax, 255
			jna LLeftSet
			mov eax, 255

LLeftSet :
		 cmp edx, 255
		 jna LRightSet
		 mov edx, 255

LRightSet :
		  and eax, 0F8h
		  add esi, 20
		  and edx, 0F8h
		  mov edi, ds : dword ptr[16 + ebx]
		  mov ecx, ds : dword ptr[12 + 16 + esp]
		  add esi, edi
		  shl eax, 7
		  add edi, ecx
		  shl edx, 7
		  mov ds : dword ptr[16 + ebx], edi
		  add eax, offset snd_scaletable
		  add edx, offset snd_scaletable
		  sub ebx, ebx
		  mov bl, ds : byte ptr[-1 + esi + ecx * 1]
		  test ecx, 1
		  jz LMix8Loop
		  mov edi, ds : dword ptr[eax + ebx * 4]
		  mov ebp, ds : dword ptr[edx + ebx * 4]
		  add edi, ds : dword ptr[paintbuffer + 0 - 8 + ecx * 8]
		  add ebp, ds : dword ptr[paintbuffer + 4 - 8 + ecx * 8]
		  mov ds : dword ptr[paintbuffer + 0 - 8 + ecx * 8], edi
		  mov ds : dword ptr[paintbuffer + 4 - 8 + ecx * 8], ebp
		  mov bl, ds : byte ptr[-2 + esi + ecx * 1]
		  dec ecx
		  jz LDone

LMix8Loop :
		mov edi, ds : dword ptr[eax + ebx * 4]
			mov ebp, ds : dword ptr[edx + ebx * 4]
			add edi, ds : dword ptr[paintbuffer + 0 - 8 + ecx * 8]
			add ebp, ds : dword ptr[paintbuffer + 4 - 8 + ecx * 8]
			mov bl, ds : byte ptr[-2 + esi + ecx * 1]
			mov ds : dword ptr[paintbuffer + 0 - 8 + ecx * 8], edi
			mov ds : dword ptr[paintbuffer + 4 - 8 + ecx * 8], ebp
			mov edi, ds : dword ptr[eax + ebx * 4]
			mov ebp, ds : dword ptr[edx + ebx * 4]
			mov bl, ds : byte ptr[-3 + esi + ecx * 1]
			add edi, ds : dword ptr[paintbuffer + 0 - 8 * 2 + ecx * 8]
			add ebp, ds : dword ptr[paintbuffer + 4 - 8 * 2 + ecx * 8]
			mov ds : dword ptr[paintbuffer + 0 - 8 * 2 + ecx * 8], edi
			mov ds : dword ptr[paintbuffer + 4 - 8 * 2 + ecx * 8], ebp
			sub ecx, 2
			jnz LMix8Loop

LDone :
		pop ebp
			pop ebx
			pop edi
			pop esi
			ret
	}
}


void SND_PaintChannelFrom16 (channel_t *ch, sfxcache_t *sc, int count)
{
	int data;
	int left, right;
	int leftvol, rightvol;
	signed short *sfx;
	int	i;

	leftvol = ch->leftvol;
	rightvol = ch->rightvol;
	sfx = (signed short *) sc->data + ch->pos;

	for (i = 0; i < count; i++)
	{
		data = sfx[i];
		left = (data * leftvol) >> 8;
		right = (data * rightvol) >> 8;
		paintbuffer[i].left += left;
		paintbuffer[i].right += right;
	}

	ch->pos += count;
}


void S_PaintChannels (int endtime)
{
	int 	i;
	int 	end;
	channel_t *ch;
	sfxcache_t	*sc;
	int		ltime, count;

	while (paintedtime < endtime)
	{
		// if paintbuffer is smaller than DMA buffer
		end = endtime;
		if (endtime - paintedtime > PAINTBUFFER_SIZE)
			end = paintedtime + PAINTBUFFER_SIZE;

		// clear the paint buffer
		Q_memset (paintbuffer, 0, (end - paintedtime) * sizeof (portable_samplepair_t));

		// paint in the channels.
		ch = channels;
		for (i = 0; i < total_channels; i++, ch++)
		{
			if (!ch->sfx)
				continue;
			if (!ch->leftvol && !ch->rightvol)
				continue;
			sc = S_LoadSound (ch->sfx);
			if (!sc)
				continue;

			ltime = paintedtime;

			while (ltime < end)
			{	// paint up to end
				if (ch->end < end)
					count = ch->end - ltime;
				else
					count = end - ltime;

				if (count > 0)
				{
					if (sc->width == 1)
						SND_PaintChannelFrom8 (ch, sc, count);
					else
						SND_PaintChannelFrom16 (ch, sc, count);

					ltime += count;
				}

				// if at end of loop, restart
				if (ltime >= ch->end)
				{
					if (sc->loopstart >= 0)
					{
						ch->pos = sc->loopstart;
						ch->end = ltime + sc->length - ch->pos;
					}
					else
					{	// channel just stopped
						ch->sfx = NULL;
						break;
					}
				}
			}

		}

		// transfer out according to DMA format
		S_TransferPaintBuffer (end);
		paintedtime = end;
	}
}

void SND_InitScaletable (void)
{
	int		i, j;

	for (i = 0; i < 32; i++)
		for (j = 0; j < 256; j++)
			snd_scaletable[i][j] = ((signed char) j) * i * 8;
}


